#!/usr/bin/python
#-*- coding: utf-8 -*-
################################################################################
# Vérification des conditions avant mise à niveau                              #
################################################################################
import sys
import glob
import os
import shlex
import re
import itertools
import operator
from creole.fonctionseole import get_eole_kernel, get_wanted_kernel
from pyeole.ihm import print_orange, print_red
from pyeole.process import system_out
from ConfigParser import SafeConfigParser
import apt
import apt_pkg
try:
    from creole import parsedico
except ImportError as e:
    print "Impossible d'exécuter le script : {0}".format(e)
    sys.exit(1)

release_path = '/etc/eole/release'
with open(release_path, 'r') as release_file:
    release = dict([line.split('=') for line in release_file.readlines()])
    module = release['EOLE_MODULE'].strip()
ROOT_PACKAGE = "all-{0}".format(module)

apt_pkg.init_config()
apt_pkg.init_system()
cache = apt_pkg.Cache()

TITLE_FORMATS = {1: {'template': u'{0}\n{1}\n{0}\n\n',
                     'marker': u'='},
                 2: {'template': u'\n{1}\n{0}\n\n',
                     'marker': u'-'},
                 3: {'template': u'{1}\n{0}\n\n',
                     'marker': u'\''}}

OPERATOR_MAPPING = {'=': {'dir': operator.eq,
                          'inv': operator.ne},
                    '>': {'dir': operator.gt,
                          'inv': operator.lt},
                    '<': {'dir': operator.lt,
                          'inv': operator.gt},
                    '<=': {'dir': operator.le,
                           'inv': operator.ge},
                    '=<': {'dir': operator.le,
                           'inv': operator.ge},
                    '>=': {'dir': operator.ge,
                           'inv': operator.le},
                    '=>': {'dir': operator.ge,
                           'inv': operator.le}}


def make_rst_title(level, content):
    """
    Return title formatted for rst output
    :param level: level conditionning markers
    :type level: int
    :param content: title to be formatted
    :type content: unicode
    """
    global TITLE_FORMATS
    formatter = TITLE_FORMATS[level]
    return formatter['template'].format(formatter['marker']*len(content),
                                        content)


def int_if_possible(value):
    """
    Return int if coercion is possible
    :param value: value to coerce to int
    :type value: str
    """
    try:
        value = int(value)
    except:
        pass
    return value


def split_version(version):
    """
    return version as list splitting subnumbers
    :param version: version number string
    :type version: string
    """
    version_splitted = re.split('[-\.]', version)
    version_splitted = map(int_if_possible, version_splitted)
    return version_splitted


def get_packages_lists(rep):
    """
    Return conf with package and version to control
    :param rep: path of the files containing packages list
    :type rep: str
    """
    sections = ['blacklist', 'whitelist', 'hp_device', 'dell_device']
    files = glob.glob(os.path.join(rep,'*.ini'))
    conf = SafeConfigParser()
    conf.read(files)
    for section in sections:
        if not conf.has_section(section):
            conf.add_section(section)
    return conf


def debfoster(package):
    """
    Return debfoster result for package.
    :param package: package we want recursive dependencies of
    :type package: str
    """
    deps = []
    deps_re = re.compile(r'(?P<package>[\w.+-]*)\s*(\((?P<constraint>[=<>]+) (?P<version>.*)\))*')
    cmd = shlex.split('debfoster -n -d {0}'.format(package))
    code, res, err = system_out(cmd)
    if code == 0:
        res = ' '.join(p for p in res.split('\n')[1:])
        deps = [(deps_re.search(reg).group('package'),(deps_re.search(reg).group('constraint'), deps_re.search(reg).group('version'))) for reg in res.split()]
    return deps

def get_installed_packages():
    """
    Return installed packages
    """
    global cache
    return [p for p in cache.packages if p.current_state == apt_pkg.CURSTATE_INSTALLED]


def test_version(candidate_package, installed_package, dir='dir'):
    """
    Return True if installed package respect version of candidate package
    :param candidate_package: package with information on version
    :type candidate_package: tuple (name, (operator, version))
    :param installed_package: installed package
    :type installed_package: apt_pkg.Package
    :param dir: wether comparaison operator is direct or inverse
    :type dir: 'dir'|'inv'
    """
    ver_re = re.compile(r'(?P<operator>[=\<\>]+)\s*(?P<version>.*)')
    ver_installed = split_version(installed_package.current_ver.ver_str)
    if candidate_package[1] != '':
        try:
            op_candidate = ver_re.search(candidate_package[1]).group('operator')
            ver_candidate = split_version(ver_re.search(candidate_package[1]).group('version'))
            test = OPERATOR_MAPPING[op_candidate][dir](ver_candidate, ver_installed)
        except AttributeError:
            test = True
    else:
        test = True
    return test


def installed_packages_from(installed_packages, packages, strict=False, dir='dir'):
    """
    Return subset of installed packages.
    :param installed_packages: list of installed packages
    :type installed_packages: list of apt_pkg.Package
    :param packages: list of packages
    :type packages: list of tuples
    :param strict: wether or not to compare version
    :type strict: boolean
    :param dir: wether comparaison operator is direct or inverse
    :type dir: 'dir'|'inv'
    """
    global cache
    filtered_packages = []
    for package in packages:
        if package[0] in [p.name for p in installed_packages]:
            if not strict:
                filtered_packages.append(cache[package[0]])
            elif test_version(package, cache[package[0]], dir):
                filtered_packages.append(cache[package[0]])
    return filtered_packages


def detect_hardware():
    """
    Return result of command lshw to gather system information
    """
    detected_hardware = {}
    cmd = shlex.split('lshw -class system')
    code, res, err = system_out(cmd)
    if code == 0:
        res = [row.split(': ') for row in res.split('\n')[1:-1]]
        for field in res:
            detected_hardware[field[0].strip()] = field[1].strip()
    else:
        detected_hardware.update({'product': u'inconnu',
                                  'vendor': u'inconnu'})
    return detected_hardware


def test_hold_packages(packages, fd):
    """
    Return report content with information on packages with hold selected status
    :param packages: list of installed packages
    :type packages: liste of apt_pkg.Package
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    test = True
    title = "Vérification des paquets à la mise à jour bloquée".decode('utf-8')
    test_report = make_rst_title(2, title)
    hold_packages = [p for p in packages if p.selected_state == apt_pkg.SELSTATE_HOLD]
    if len(hold_packages) > 0:
        test = False
        test_report += "Test échoué\n\n".decode('utf-8')
        test_report += "La mise à jour des paquets suivants est bloquée :\n\n  -".decode('utf-8')
        test_report += "\n  -".join([p.name for p in hold_packages]).decode('utf-8')
        test_report += "\n\n".decode('utf-8')
        test_report += """Débloquez la mise à jour de ces paquets
avant de procéder à l'upgrade du serveur.
S'il s'agit des paquets eole-geo-ide-geobase
ou arkeia, reportez vous à la fiche correspondante.\n\n""".decode('utf-8')
    else:
        test_report += "Test OK\n\n".decode('utf-8')
    fd.write(test_report.encode('utf-8'))
    return test


def repr_package(package):
    """
    Return str representation of package (name and version)
    :param package: package
    :type package: apt_pkg.Package
    """
    return "{0} ({1})".format(package.name, package.current_ver.ver_str)


def test_hardware(packages, packages_of_interest, fd):
    """
    Return report content with information on packages for hp or dell hardware.
    :param packages: list of installed packages
    :type packages: list of apt_pkg.Package
    :param packages_of_interest: list of packages from files
    :type packages_of_interest: SafeConfigParser with sections hp_device and dell_device
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    test = True
    title = "Vérification du matériel".decode('utf-8')
    test_report = make_rst_title(2, title)
    test_report += "Détection des paquets de supervision du matériel :\n\n".decode('utf-8')
    hp_packages = packages_of_interest.items('hp_device')
    installed_hp_packages = [repr_package(p) for p in installed_packages_from(packages, hp_packages)]
    dell_packages = packages_of_interest.items('dell_device')
    installed_dell_packages = [repr_package(p) for p in installed_packages_from(packages, dell_packages)]
    if len(installed_hp_packages) == 0 and len(installed_dell_packages) == 0:
        test_report += "Aucun paquet détecté\n\n".decode('utf-8')
        test_report += "Test OK\n".decode('utf-8')
    else:
        test = False
        detected_hardware = detect_hardware()
        hardware_str = u' - '.join([detected_hardware['vendor'],
                                       detected_hardware['product']])
        test_report += "Les paquets suivants sont installés :\n\n  -".decode('utf-8')
        test_report += "\n  -".join(itertools.chain(installed_hp_packages, installed_dell_packages)).decode('utf-8')
        test_report += "\n\n".decode('utf-8')
        test_report += "Le matériel est identifié comme : {0}\n\n".format(hardware_str).decode('utf-8')
        test_report += "Procédure de désinstallation :\n\n"
        if detected_hardware['vendor'] == 'Dell':
            test_report += """  apt-get upgrade
  apt-eole remove {0}
  rm -rf /etc/opt/dell
  rm -rf /opt/dell
  dpkg --force-all --purge {0}
  reboot du serveur et contrôle du bon démarrage des interfaces.\n\n""".format(installed_dell_packages).decode('utf-8')
        if detected_hardware['vendor'] == 'HP':
            test_report += """  apt-eole remove {0}
  dpkg --force-all --purge {0}\n\n""".format(installed_hp_packages).decode('utf-8')
        else:
            test_report += """Merci de suivre la procédure de désinstallation
  ou de lancer le script de désinstallation
  correspondant à la marque de votre serveur.\n\n""".decode('utf-8')
    fd.write(test_report.encode('utf-8'))
    return test


def test_kernel(fd):
    """
    Return report content with information on kernel.
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    test = True
    title = "Vérification de la version de noyau utilisée".decode('utf-8')
    test_report = make_rst_title(2, title)
    if get_eole_kernel()[0] == get_wanted_kernel():
        test_report += "Test OK\n\n".decode('utf-8')
    else:
        test = False
        test_report += """La mise à jour du noyau de serveur est bloquée.
Débloquez-la avant de procéder à l'upgrade du serveur\n\n""".decode('utf-8')
    fd.write(test_report.encode('utf-8'))
    return test


def test_local_dicts(fd):
    """
    Return local report
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    test = True
    subtitle = "Dictionnaires hors module et variante".decode('utf-8')
    test_report = make_rst_title(3, subtitle)
    local_dicts_glob = '/usr/share/eole/creole/dicos/local/*.xml'
    local_dicts = [os.path.basename(f) for f in glob.glob(local_dicts_glob)]
    if len(local_dicts) > 0:
        test = False
        test_report += u"  -"
        test_report += "  -".join(local_dicts).decode('utf-8')
        test_report += "\n\n".decode('utf-8')
    else:
        test_report += "Aucun dictionnaire local.\n\n".decode('utf-8')
    fd.write(test_report.encode('utf-8'))
    return test


def test_patchs(fd):
    """
    Return patchs
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    test = True
    subtitle = "Patchs".decode('utf-8')
    test_report = make_rst_title(3, subtitle)
    local_dicts_glob = '/usr/share/eole/creole/patch/*.patch'
    local_dicts = [os.path.basename(f) for f in glob.glob(local_dicts_glob)]
    if len(local_dicts) > 0:
        test = False
        test_report += u"  -"
        test_report += "  -".join(local_dicts).decode('utf-8')
        test_report += "\n\n".decode('utf-8')
    else:
        test_report += "Aucun patch local.\n\n".decode('utf-8')
    fd.write(test_report.encode('utf-8'))
    return test


def test_custom_packages(installed_packages, packages, fd):
    """
    Return packages not in module or variante.
    :param installed_packages: list of installed packages
    :type installed_packages: list of apt_pkg.Package
    :param packages: list of packages
    :type packages: list of tuples
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    test = True
    subtitle = "Paquets non directement liés au module ou à la variante".decode('utf-8')
    test_report = make_rst_title(3, subtitle)
    sip = set([p.name for p in installed_packages])
    wanted_packages = [ROOT_PACKAGE, ]
    wanted_packages.extend(packages.options('whitelist'))
    kernel_re = re.compile(r'linux-(generic|headers|image).*')
    wanted_packages.extend([p for p in sip if kernel_re.match(p)])
    deps = [debfoster(wp) for wp in wanted_packages]
    wanted_deps = set([p[0] for p in itertools.chain.from_iterable(deps)])
    wanted_packages = wanted_deps.union(set(wanted_packages))
    unwanted_packages = sip.difference(wanted_packages)
    if len(list(unwanted_packages)) > 0:
        test = False
        test_report += "Liste des paquets :\n\n  -".decode('utf-8')
        test_report += '\n  -'.join(sorted(unwanted_packages)).decode('utf-8')
        test_report += '\n\n'.decode('utf-8')
    else:
        test_report += "Aucun paquet hors module ou variante\n\n".decode('utf-8')
    blacklist = packages.items('blacklist')
    blacklisted_packages = [repr_package(p) for p in installed_packages_from(installed_packages, blacklist, strict=True, dir='inv')]
    if len(blacklisted_packages) > 0:
        test = False
        test_report += "Liste des paquets nécessitant une attention particulière :\n\n  -".decode('utf-8')
        test_report += '\n  -'.join(sorted(blacklisted_packages)).decode('utf-8')
        test_report += '\n\n'.decode('utf-8')
    else:
        test_report += "Aucun paquet nécessitant une attention particulière\n\n".decode('utf-8')
    fd.write(test_report.encode('utf-8'))
    return test


def test_customisation(installed_packages, packages, fd):
    """
    :param installed_packages: list of installed packages
    :type installed_packages: list of apt_pkg.Package
    :param packages: list of packages
    :type packages: list of tuples
    :param fd: file descriptor open in write or append mode
    :type fd: file descriptor
    """
    title = "Vérification des adaptations locales".decode('utf-8')
    fd.write(make_rst_title(2, title).encode('utf-8'))
    test_ld = test_local_dicts(fd)
    test_p = test_patchs(fd)
    test_cp =  test_custom_packages(installed_packages, packages, fd)
    return test_ld and test_p and test_cp


if __name__ == '__main__':

    print_orange("Examen des conditions préalables pour la montée de version\n")
    packages = get_packages_lists('/root/scripts/upgrade_prerequisites')
    installed_packages = get_installed_packages()
    dirty = 'debfoster' not in [p.name for p in installed_packages]
    if dirty:
        print_orange("\nInstallation du programme debfoster pour déterminer les dépendances du module (ce programme sera désinstallé à la fin de la procédure de test).")
        cmd = shlex.split('apt-get install debfoster')
        system_out(cmd)
        packages.set('whitelist', 'debfoster', '')
    report_file = '/tmp/test_report.rst'
    with open('/tmp/test_report.rst', 'w') as test_report:
        test_report.write(make_rst_title(1, "Rapport des conditions requises pour la montée de niveau".decode('utf-8')).encode('utf-8'))
    with open('/tmp/test_report.rst', 'a') as test_report:
        print "\nTest des paquets hold"
        test_hp = test_hold_packages(installed_packages, test_report)
        print "\nTest du matériel"
        test_h =  test_hardware(installed_packages, packages, test_report)
        print "\nTest du noyau"
        test_k =  test_kernel(test_report)
        print "\nTest des adaptations"
        test_c =  test_customisation(installed_packages, packages, test_report)
        if test_hp and test_h and test_k and test_c:
            message = """\nLe serveur peut être migré grâce à la commande :
  Upgrade-Auto"""
        else:
            message = "\nLe serveur nécessite des interventions avant migration."
        test_report.write(make_rst_title(1, "Bilan").encode('utf-8'))
        test_report.write(message)
    if dirty:
        cmd = shlex.split('apt-get remove -y debfoster')
        system_out(cmd)
    print_red("\n\nLe résultat des tests peut être consulté dans {0}".format(report_file))
